/***********************
   read FatFs_readme.h
 ***********************/
#ifndef FF_h
#define FF_h

#include <Arduino.h>

/*---------------------------------------------------------------------------/
  / Functions and Buffer Configurations
  /---------------------------------------------------------------------------*/
#define  _USE_STRFUNC  1  /* 0:Disable or 1-2:Enable */
/* To enable string functions, set _USE_STRFUNC to 1 or 2 (strip \r). */

/*---------------------------------------------------------------------------/
  / Drive/Volume Configurations
  /---------------------------------------------------------------------------*/
#define _FS_NOFSINFO  0  /* 0 to 3 */
/* If you need to know correct free space on the FAT32 volume, set bit 0 of this
  /  option and f_getfree() function at first time after volume mount will force
  /  a full FAT scan. Bit 1 controls the last allocated cluster number as bit 0.
  /
  /  bit0=0: Use free cluster count in the FSINFO if available.
  /  bit0=1: Do not trust free cluster count in the FSINFO.
  /  bit1=0: Use last allocated cluster number in the FSINFO if available.
  /  bit1=1: Do not trust last allocated cluster number in the FSINFO.
*/

/*---------------------------------------------------------------------------/
  / System Configurations
  /---------------------------------------------------------------------------*/
#define  _FS_LOCK  2  /* 0:Disable or >=1:Enable */
/* To enable file lock control feature, set _FS_LOCK to 1 or greater.
  /  The value defines how many files/sub-directories can be opened simultaneously.
  /  This feature consumes _FS_LOCK * 12 bytes of bss area. */

/* Definitions of volume management */

#define LD2PD(vol) (byte)(vol)  /* Each logical drive is bound to the same physical drive number */
#define LD2PT(vol) 0  /* Find first valid partition or in SFD */

typedef char TCHAR;

/* File system object structure (FATFS) */

struct FATFS {
  uint8_t  fs_type;     /* FAT sub-type (0:Not mounted) */
  uint8_t  drv;         /* Physical drive number */
  uint8_t  csize;       /* Sectors per cluster (1,2,4...128) */
  uint8_t  n_fats;      /* Number of FAT copies (1 or 2) */
  uint8_t  wflag;       /* win[] flag (b0:dirty) */
  uint8_t  fsi_flag;    /* FSINFO flags (b7:disabled, b0:dirty) */
  uint16_t id;          /* File system mount ID */
  uint16_t n_rootdir;   /* Number of root directory entries (FAT12/16) */
  uint32_t last_clust;  /* Last allocated cluster */
  uint32_t free_clust;  /* Number of free clusters */
  uint32_t n_fatent;    /* Number of FAT entries (= number of clusters + 2) */
  uint32_t fsize;       /* Sectors per FAT */
  uint32_t volbase;     /* Volume start sector */
  uint32_t fatbase;     /* FAT start sector */
  uint32_t dirbase;     /* Root directory start sector (FAT32:Cluster#) */
  uint32_t database;    /* Data start sector */
  uint32_t winsect;     /* Current sector appearing in the win[] */
  uint8_t  win[512];    /* Disk access window for Directory, FAT (and file data at tiny cfg) */
};

/* File object structure (FIL) */

struct FIL {
  FATFS*   fs;        /* Pointer to the related file system object (**do not change order**) */
  uint16_t id;        /* Owner file system mount ID (**do not change order**) */
  uint8_t  flag;      /* File status flags */
  uint8_t  err;       /* Abort flag (error code) */
  uint32_t fptr;      /* File read/write pointer (Zeroed on file open) */
  uint32_t fsize;     /* File size */
  uint32_t sclust;    /* File data start cluster (0:no data cluster, always 0 when fsize is 0) */
  uint32_t clust;     /* Current cluster of fpter */
  uint32_t dsect;     /* Current data sector of fpter */
  uint32_t dir_sect;  /* Sector containing the directory entry */
  byte*    dir_ptr;   /* Pointer to the directory entry in the window */
#if _FS_LOCK
  uint16_t lockid;    /* File lock ID (index of file semaphore table Files[]) */
#endif
};

/* Directory object structure (DIR) */

struct DIR {
  FATFS*   fs;      /* Pointer to the owner file system object (**do not change order**) */
  uint16_t id;      /* Owner file system mount ID (**do not change order**) */
  uint16_t index;   /* Current read/write index number */
  uint32_t sclust;  /* Table start cluster (0:Root dir) */
  uint32_t clust;   /* Current cluster */
  uint32_t sect;    /* Current sector */
  byte*    dir;     /* Pointer to the current SFN entry in the win[] */
  byte*    fn;      /* Pointer to the SFN (in/out) {file[8],ext[3],status[1]} */
#if _FS_LOCK
  uint16_t lockid;  /* File lock ID (index of file semaphore table Files[]) */
#endif
};

/* File status structure (FILINFO) */

struct FILINFO {
  uint32_t fsize;      /* File size */
  uint16_t fdate;      /* Last modified date */
  uint16_t ftime;      /* Last modified time */
  uint8_t  fattrib;    /* Attribute */
  TCHAR    fname[13];  /* Short file name (8.3 format) */
};

/* File function return code (FRESULT) */

enum FRESULT {
  FR_OK = 0,               /* (0) Succeeded */
  FR_DISK_ERR,             /* (1) A hard error occurred in the low level disk I/O layer */
  FR_INT_ERR,              /* (2) Assertion failed */
  FR_NOT_READY,            /* (3) The physical drive cannot work */
  FR_NO_FILE,              /* (4) Could not find the file */
  FR_NO_PATH,              /* (5) Could not find the path */
  FR_INVALID_NAME,         /* (6) The path name format is invalid */
  FR_DENIED,               /* (7) Access denied due to prohibited access or directory full */
  FR_EXIST,                /* (8) Access denied due to prohibited access */
  FR_INVALID_OBJECT,       /* (9) The file/directory object is invalid */
  FR_WRITE_PROTECTED,      /* (10) The physical drive is write protected */
  FR_INVALID_DRIVE,        /* (11) The logical drive number is invalid */
  FR_NOT_ENABLED,          /* (12) The volume has no work area */
  FR_NO_FILESYSTEM,        /* (13) There is no valid FAT volume */
  FR_MKFS_ABORTED,         /* (14) The f_mkfs() aborted due to any parameter error */
  FR_TIMEOUT,              /* (15) Could not get a grant to access the volume within defined period */
  FR_LOCKED,               /* (16) The operation is rejected according to the file sharing policy */
  FR_NOT_ENOUGH_CORE,      /* (17) LFN working buffer could not be allocated */
  FR_TOO_MANY_OPEN_FILES,  /* (18) Number of open files > _FS_SHARE */
  FR_INVALID_PARAMETER     /* (19) Given parameter is invalid */
};

#define f_eof(fp)   (((fp)->fptr == (fp)->fsize) ? 1 : 0)
#define f_error(fp) ((fp)->err)
#define f_tell(fp)  ((fp)->fptr)
#define f_size(fp)  ((fp)->fsize)

#define EOF (-1)

/*--------------------------------------------------------------*/
/* Additional user defined functions  */

/* RTC function */
uint32_t get_fattime (void);

/*--------------------------------------------------------------*/
/* Flags and offset address  */

/* File access control and file status flags (FIL.flag) */

#define FA_READ  0x01
#define FA_OPEN_EXISTING  0x00
#define FA_WRITE  0x02
#define FA_CREATE_NEW  0x04
#define FA_CREATE_ALWAYS  0x08
#define FA_OPEN_ALWAYS  0x10
#define FA__WRITTEN  0x20
#define FA__DIRTY  0x40

/* FAT sub type (FATFS.fs_type) */

#define FS_FAT12  1
#define FS_FAT16  2
#define FS_FAT32  3

/* File attribute bits for directory entry */

#define AM_RDO  0x01   /* Read only */
#define AM_HID  0x02   /* Hidden */
#define AM_SYS  0x04   /* System */
#define AM_VOL  0x08   /* Volume label */
#define AM_LFN  0x0F   /* LFN entry */
#define AM_DIR  0x10   /* Directory */
#define AM_ARC  0x20   /* Archive */
#define AM_MASK  0x3F  /* Mask of defined bits */

/* Fast seek feature */
#define CREATE_LINKMAP  0xFFFFFFFF

/*--------------------------------*/
/* Multi-byte word access macros  */

#define LD_uShort(ptr)  (uint16_t)(((uint16_t)*((byte*)(ptr)+1)<<8)|(uint16_t)*(byte*)(ptr))
#define LD_uLong(ptr)  (uLong)(((uLong)*((byte*)(ptr)+3)<<24)|((uLong)*((byte*)(ptr)+2)<<16)|((uint16_t)*((byte*)(ptr)+1)<<8)|*(byte*)(ptr))
#define ST_uShort(ptr,val)  *(byte*)(ptr)=(byte)(val); *((byte*)(ptr)+1)=(byte)((uint16_t)(val)>>8)
#define ST_uLong(ptr,val)  *(byte*)(ptr)=(byte)(val); *((byte*)(ptr)+1)=(byte)((uint16_t)(val)>>8); *((byte*)(ptr)+2)=(byte)((uLong)(val)>>16); *((byte*)(ptr)+3)=(byte)((uLong)(val)>>24)

class SdFatFs {
  public:
    FRESULT f_open (FIL* fp, const TCHAR* path, uint8_t mode);                /* Open or create a file */
    FRESULT f_close (FIL* fp);                                                /* Close an open file object */
    FRESULT f_read (FIL* fp, void* buff, uint16_t btr, uint16_t* br);         /* Read data from a file */
    FRESULT f_write (FIL* fp, const void* buff, uint16_t btw, uint16_t* bw);  /* Write data to a file */
    FRESULT f_lseek (FIL* fp, uint32_t ofs);                                  /* Move file pointer of a file object */
    FRESULT f_sync (FIL* fp);                                                 /* Flush cached data of a writing file */
    FRESULT f_mount (FATFS* fs, const TCHAR* path, uint8_t opt);              /* Mount/Unmount a logical drive */
    int     f_putc (TCHAR c, FIL* fp);  				                              /* Put a character to the file */
    int     f_puts (const TCHAR* str, FIL* cp);                               /* Put a string to the file */
    int     f_printf (FIL* fp, const TCHAR* str, ...);                        /* Put a formatted string to the file */
    TCHAR*  f_gets (TCHAR* buff, int len, FIL* fp);                           /* Get a string from the file */
};

#endif




